# Rcon::HL2 - Perl extension Half-Life 2 (Source) engine Rcon interface
#
# $Id:$
#

package Rcon::HL2;

use strict;
use warnings;
use IO::Socket;
use IO::Select;

# release version
our $VERSION = "0.02";

# constants
sub TRUE { 1 }
sub CMD  { 2 }
sub AUTH { 3 }

sub new {
	my $class = shift;

	# create object
	my $self = {};
	bless($self, $class);

	# parse constructor args
	my %attr = @_;
	while (my ($key, $val) = each %attr) {
		$key = lc($key);
		if    ($key eq "hostname") { $self->hostname($val) }
		elsif ($key eq "port")     { $self->port($val)     }
		elsif ($key eq "password") { $self->password($val) }
		elsif ($key eq "timeout")  { $self->timeout($val)  }
		else { print STDERR "uknown attribute: $key\n" }
	}

	return $self;
}

sub authenticated {
	my $self = shift;
	my $authenticated = shift;
	my $obj = \$self->{authenticated};
	if (defined $authenticated) {
		$$obj = $authenticated;
		return $self;
	} else {
		return $$obj;
	}
}

sub password {
	my $self = shift;
	my $password = shift;
	my $obj = \$self->{password};
	if (defined $password) {
		$$obj = $password;
		return $self;
	} else {
		return $$obj;
	}
}

sub hostname {
	my $self = shift;
	my $hostname = shift;
	my $obj = \$self->{hostname};
	if (defined $hostname) {
		$$obj = $hostname;
		return $self;
	} else {
		return $$obj;
	}
}

sub port {
	my $self = shift;
	my $port = shift;
	my $obj = \$self->{port};
	if (defined $port) {
		$$obj = $port;
		return $self;
	} else {
		if (!defined $$obj) {
			$$obj = 27015;
		}

		return $$obj;
	}
}

sub sequence {
	my $self = shift;
	my $sequence = shift;
	my $obj = \$self->{sequence};
	if (defined $sequence) {
		$$obj = $sequence;
		return $self;
	} else {
		if (!defined $$obj) {
			$$obj = 0;
		} else {
			$$obj++;
		}

		return $$obj;
	}
}

sub socket {
	my $self = shift;
	my $socket = shift;
	my $obj = \$self->{socket};
	if (defined $socket) {
		$$obj = $socket;
		return $self;
	} else {
		if (!defined $$obj) {
			$$obj = $self->connectToServer();
		}

		return $$obj;
	}
}

sub pollInterval {
	my $self = shift;
	my $pollInterval = shift;
	my $obj = \$self->{pollInterval};
	if (defined $pollInterval) {
		$$obj = $pollInterval;
		return $self;
	} else {
		if (!defined $$obj) {
			$$obj = 0.25;
		}

		return $$obj;
	}
}

sub timeout {
	my $self = shift;
	my $timeout = shift;
	my $obj = \$self->{timeout};
	if (defined $timeout) {
		$$obj = $timeout;
		return $self;
	} else {
		if (!defined $$obj) {
			$$obj = 3;
		}

		return $$obj;
	}
}

sub connectToServer {
	my $self = shift;

	my $socket   = IO::Socket::INET->new(
		PeerAddr        => $self->hostname(),
		PeerPort        => $self->port(),
		Timeout		=> $self->timeout(),
		Proto           => "tcp",
		Type            => SOCK_STREAM,
	) || die "failed to connect: $!\n";

	return $socket;
}

# protocol decoded, roughly:
#(length of pass+10)    0*3 (02)   0*3 03 0*3 (pass) 0*2 
#(length of command+10) 0*3 (seq#) 0*3 02 0*3 (cmd) 0*2
#
# more precisely:
#
# [2 int payload size] [payload] [2 nulls to terminate]
# command payloads have 10 bytes of metadeta which the
# packet() method sets up for you.  i have not decoded
# the meaning of the entire payload, but i got the
# essentials required to authenticate and send
# rcon commands.  maybe the SDK has the precise
# meaning off the command metadata.

my $sep = "\x00" x 3;
my $end = "\x00" x 2;
sub packet {
	my $self     = shift;
	my $type     = shift;
	my $payload  = shift;

	my $sequence;
	if ($type == AUTH) {
		$sequence = 2;
	} else {
		$sequence = $self->sequence();
	}

	my $packet = sprintf(
		"%s$sep$sequence$sep%s$sep$payload$end",
		chr(length($payload) + 10),
		chr($type),
	);

	return $packet;
}

sub authenticate {
	my $self = shift;

	# send authentication packet to server
	my $socket = $self->socket();
	print $socket $self->packet(AUTH, $self->password());

	# check response XXX: need to parse auth response..
	# why does it return 2 payloads in new update? sigh.
	my $response = $self->response() . $self->response();

	my $status;
	if (1) {
		$status = 1;
	} else {
		printf STDERR "failed authentication!\n";
		$status = 0;
	}

	$self->authenticated($status);
	return $status;
}

sub run {
	my $self = shift;
	my $command = shellwords(@_);

	if (!$self->authenticated()) {
		my $status = $self->authenticate();
		return unless $status;
	}

	my $socket = $self->socket();
	print $socket $self->packet(CMD, $command);
}

sub response {
	my $self = shift;

	my $socket = $self->socket();
	my $timeout = $self->timeout();
	my $pollInterval = $self->pollInterval();
	my $select = IO::Select->new($socket);

	my $payloadSize;
	my $payloadFirstInt;
	my $payload = "";
	my $termSequence = 0;
	my $start = time();
	while (TRUE) {
		# timeout exceeded
		last if ((time() - $start) > $timeout);

		# poll socket
		my @ready = $select->can_read($pollInterval);

		# exit loop if we lose the connection
		last unless $socket->connected();

		# loop if we did not read anything
		next unless @ready;

		# read a byte from the socket
		my $read = $socket->sysread(my $chr, 1, 0);

		# continue loop if we failed to read
		next unless defined $read;

		# reading 0 bytes means eof
		last if ($read <= 0);

		# parse packet
		# (two byte payload size integer)(payload)(2 null terminators)
		if (!defined $payloadFirstInt) {
			$payloadFirstInt = ord($chr);
		} elsif (!defined $payloadSize) {
			$payloadSize = ($payloadFirstInt + (ord($chr) * 256));
		} elsif (length($payload) < $payloadSize) {
			$payload .= $chr;
		} elsif ($chr eq "\x00") {
			$termSequence++;
			last if ($termSequence == 2);
		}
	}

	return $payload;
}

# adapted from shellwords.pl in perl distribution.
# this takes a list or string of commands and
# parses them like a shell (e.g. honors quoting), and
# then quotes each arg for rcon compliance. e.g.:
#
# shellwords(qq/changelevel de_dust/) becomes "changelevel" "de_dust"
# shellwords(qq/say "this is a test"/) becomes "say" "this is a test"
sub shellwords {
	my $text = join('', @_);
	return if (!defined $text);

	# strip leading whitespace
	$text =~ s/^\s+//;

	# fix deep recursion case (can't escape newline :P)
	# found by hlprmnky
	if ($text =~ /(?:[^\\]|^)\x5c$/) {
		print STDERR "Unmatched escape\n";
		return undef;
	}

	my @words;
	while ($text ne "") {
		my $field;
		while (TRUE) {
			my $snippet;
			if ($text =~ s/^"(([^"\\]|\\.)*)"//) {
				($snippet = $1) =~ s#\\(.)#$1#g;
			} elsif ($text =~ /^"/) {
				print STDERR "Unmatched double quote\n";
				return;
			} elsif ($text =~ s/^'(([^'\\]|\\.)*)'//) {
				($snippet = $1) =~ s#\\(.)#$1#g;
			} elsif ($text =~ /^'/) {
				print STDERR "Unmatched single quote\n";
				return;
			} elsif ($text =~ s/^\\(.)//) {
				$snippet = $1;
			} elsif ($text =~ s/^([^\s\\'"]+)//) {
				$snippet = $1;
			} else {
				$text =~ s/^\s+//;
				last;
			}

			$field .= $snippet;
		}
		$field = qq("$field");
		push(@words, $field);
	}
	return join(" ", @words);
}

1;

__END__

=head1 NAME

Rcon::HL2 - Perl extension Half-Life 2 (Source) engine Rcon interface

=head1 SYNOPSIS

  use Rcon::HL2;
  my $rcon = Rcon::HL2->new(
  	hostname => "insub.org",
	password => "yourpass",
	timeout  => 3,
  );

  $rcon->run("status");
  print $rcon->response();
  $rcon->timeout(5);
  $rcon->run("changelevel de_dust");

=head1 DESCRIPTION

Use this module to send "rcon" (remote control) commands to a
Source server, such as Counter-Strike Source.

=head1 METHODS

=over 4

=item $rcon = Rcon::HL2->new()

Create a new rcon object.  You can specify the hostname,
password and/or timeout in the constructor, or use the class
methods to change them (see SYNOPSIS).

=item $rcon->authenticated()

Returns true if session has succesfully authenticated.

=item $rcon->password()

Returns current password, or sets it.  Note that setting
this after authentication will not have any effect unless
you reconnect with $rcon->authenticated(0).

=item $rcon->hostname()

Returns current hostname, or sets it.

=item $rcon->port()

Returns current port, or sets it.  Defaults to 27015.

=item $rcon->sequence()

Returns the current command sequence.  This starts
at 0 and increases with each call.  You probably
shouldn't use this in the middle of a session.

=item $rcon->socket()

Returns the IO::Socket object for the session or
creates a new one if none exists.

=item $rcon->pollInterval()

Returns interval for socket select(), or sets it.
You probably don't need to change this. Defaults to 0.25.

=item $rcon->timeout()

Returns the TCP response timeout, or sets it.  Defaults
to 3.  Note that currently certain server responses which
are unparseable will cause the response() method to hang
until the timeout is reached, so you should not set this
too high.

=item $rcon->connectToServer()

Creates and returns an IO::Socket object.

=item $rcon->packet($type, $payload)

Creats a packet to send to the remote server.
Type should be either CMD or AUTH, e.g.:

  print $socket $rcon->packet(AUTH, $rcon->password())

=item $rcon->authenticate()

Authenticates with the rcon server.  This is done automatically
when you try to run a command.

=item $rcon->run($command)

Runs a command on the remote server.

=item $response = $rcon->response()

Reads a response packet from the server.  Some packets that
are malformed (or not understood by this code) will result
in it hanging until the timeout is reached, so be sure not
to set the timeout value too high (it currently defaults to 3 seconds).

=back

=head1 CAVEATS

This module DOES NOT DO ANY COMMAND VALIDATION.  You are responsible
for sending sane commands to the server.  If you use this with CGI that
allows internet users to submit console commands, you MUST taint-check
this.  Users with RCON access can send anything to the console.  I highly recommend
that you restrict what console commands a user can send.  I really can't stress
this enough, so I'll stop here.

=head1 BUGS

I reverse engineered the rcon protocol using tcpdump and ethereal.
The nuts and bolts were pretty obvious, but there are some response
packets that response() does not parse correctly, which results in
a timeout.  User contributions are welcome.

As of this writing, there are some bugs with the rcon server itself.
One such bug is that some output goes to the console instead of to
the rcon client.  For example, the command "listid" causes the list
of banned users to spew to the physical console instead of back to
the rcon client, making it effectively useless.  If you are not getting
back a response you expected, please verify that it's not going to
the console (run srcds in screen so you can access it) before submitting
a bug report to me about it.  Or better yet, submit a bug report to Valve.

=head1 SEE ALSO

 http://gruntle.org/projects/
 http://insub.org/cs/

=head1 AUTHOR

Chris Jones, E<lt>cjones@gruntle.orgE<gt>

=head1 COPYRIGHT AND LICENSE

 Copyright (c) 2003-2004 Chris Jones <cjones@gruntle.org>
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 
 THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

=cut
